"""
PBIL (Population-Based Incremental Learning)
Salustowicz and Schmidhuber (1997, 1998)
H6hfeld and Rudolph (1997)
Berny (2000a)
Juels (1997)

Population-based incremental learning: A method for integrating genetic search based function optimization and competitive learning
Genetic algorithms and explicit search statistics
Removing the genetics from standard genetic algorithm
Fast probabilistic modeling for combinatorial optimization
Extending population-based incremental learning to continuous search spaces
"""
