"""PyPop7 is a Pure-PYthon library of POPulation-based OPtimization for single-objective, real-parameter, black-box
    problems (currently actively developed). Its main goal is to provide a unified interface and elegant implementations
    for Black-Box Optimization (BBO), particularly population-based optimizers, in order to facilitate research
    repeatability and also real-world applications.

    More specifically, for alleviating the notorious curse of dimensionality of BBO (almost based on iterative
    sampling), the primary focus of PyPop7 is to cover their State-Of-The-Art (SOTA) implementations for
    Large-Scale Optimization (LSO), though many of their other versions and variants are also included here
    (for benchmarking/mixing purpose, and sometimes even for practical purpose).
"""
